using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using BlueWest.EfMethods.EfMethods.AttributeSources;
using BlueWest.EfMethods.Extensions;
using BlueWest.EfMethods.Sources;
using BlueWest.Sources;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp.Syntax;

namespace BlueWest.EfMethods
{
    /// <summary>
    /// Ef methods source generator
    /// </summary>
    [Generator]
    public class EfMethodsGenerator: ISourceGenerator
    {
         /// <inheritdoc />
        public void Initialize(GeneratorInitializationContext context)
        {
            context.RegisterForSyntaxNotifications(() => new EfMethodsSyntaxReceiver());
        }

        /// <inheritdoc />
        public void Execute(GeneratorExecutionContext context)
        {

            //SpinWait.SpinUntil(() => Debugger.IsAttached);
            
            try
            {
                var options = SourceGenerationOptions.From(context);

                var compilation = context.Compilation
                    .AddSource(ref context, EfGeneratorAttributeSource.Generate(options))
                    .AddSource(ref context, EfAddMethodsAttributeSource.Generate(options))
                    .AddSource(ref context, EfUpdateMethodsAttributeSource.Generate(options))
                    .AddSource(ref context, EfGetOneByAttributeSource.Generate(options))
                    .AddSource(ref context, EfGetOneAttributeSource.Generate(options))
                    .AddSource(ref context, EfGetManyAttributeSource.Generate(options))
                    .AddSource(ref context, EfGetListAttributeSource.Generate(options))
                    .AddSource(ref context, EfAddToListAttributeSource.Generate(options))
                    .AddSource(ref context, EfGetOneFromListAttributeSource.Generate(options));
                    
                


                if (context.SyntaxReceiver is EfMethodsSyntaxReceiver receiver && receiver.CandidateTypes.Any())
                {
                    AddGeneratedExtensions(context, compilation, receiver.CandidateTypes, options);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                throw;
            }
        }

        private static EfTemplates GetAvailableTemplates(GeneratorExecutionContext context)
        {
            string addSourceTemplate = GetCsxMethodTemplate(context, "AddToTemplate");
            string updateSourceTemplate = GetCsxMethodTemplate(context, "UpdateTemplate");
            string getOneByTemplate = GetCsxMethodTemplate(context, "GetOneByTemplate");
            string getOneTemplate = GetCsxMethodTemplate(context, "GetOneTemplate");
            string getManyWithTemplate = GetCsxMethodTemplate(context, "GetManyTemplate");
            string getListTemplate = GetCsxMethodTemplate(context, "GetListTemplate");
            string addToListTemplate = GetCsxMethodTemplate(context, "AddToListTemplate");
            string getOneFromListTemplate = GetCsxMethodTemplate(context, "GetOneFromListTemplate");
            

            return new EfTemplates(
                addSourceTemplate, 
                updateSourceTemplate, 
                getOneByTemplate, 
                getOneTemplate, 
                getManyWithTemplate, 
                getListTemplate,
                addToListTemplate,
                getOneFromListTemplate
            );
        }

        private static void AddGeneratedExtensions(GeneratorExecutionContext context, Compilation compilation, IEnumerable<TypeDeclarationSyntax> candidateMembers, SourceGenerationOptions options)
        {
            var templates = GetAvailableTemplates(context);

            foreach (var candidateMember in candidateMembers)
            {
         
                var mappingContext = EfGeneratorContext.Create(compilation, options, candidateMember);
                
                mappingContext.Diagnostics.ForEach(context.ReportDiagnostic);
                
                if (mappingContext.Model is null)
                {
                    continue;
                }
                
                var (source, hintName) = candidateMember switch
                {
                    ClassDeclarationSyntax => EfMethodsSource.Generate(
                        mappingContext.Model, 
                        templates),
                    _ => throw new ArgumentOutOfRangeException()
                };

                context.AddSource(hintName, source);
            }
        }

        private static string GetCsxMethodTemplate(GeneratorExecutionContext context, string templateName)
        {
            return context
                .AdditionalFiles
                    .FirstOrDefault(x => x.Path.Contains(templateName))?
                    .GetText()?
                    .ToString() ?? string.Empty;
        }

    }
}